@@ -1,12 +1,13 @@
 from __future__ import absolute_import, print_function, division
-from pony.py23compat import PY2, izip, xrange, PY37, PYPY
+from pony.py23compat import PY37, PYPY, PY310
 
 import sys, types, inspect
 from opcode import opname as opnames, HAVE_ARGUMENT, EXTENDED_ARG, cmp_op
 from opcode import hasconst, hasname, hasjrel, haslocal, hascompare, hasfree
 from collections import defaultdict
 
-from pony.thirdparty.compiler import ast, parse
+#from pony.thirdparty.compiler import ast, parse
+import ast
 
 from pony.utils import throw, get_codeobject_id
 
@@ -21,15 +22,16 @@ ast_cache = {}
 def decompile(x):
     cells = {}
     t = type(x)
-    if t is types.CodeType: codeobject = x
-    elif t is types.GeneratorType: codeobject = x.gi_frame.f_code
+    if t is types.CodeType:
+        codeobject = x
+    elif t is types.GeneratorType:
+        codeobject = x.gi_frame.f_code
     elif t is types.FunctionType:
-        codeobject = x.func_code if PY2 else x.__code__
-        if PY2:
-            if x.func_closure: cells = dict(izip(codeobject.co_freevars, x.func_closure))
-        else:
-            if x.__closure__: cells = dict(izip(codeobject.co_freevars, x.__closure__))
-    else: throw(TypeError)
+        codeobject = x.__code__
+        if x.__closure__:
+            cells = dict(zip(codeobject.co_freevars, x.__closure__))
+    else:
+        throw(TypeError)
     key = get_codeobject_id(codeobject)
     result = ast_cache.get(key)
     if result is None:
@@ -38,27 +40,56 @@ def decompile(x):
         ast_cache[key] = result
     return result + (cells,)
 
+
 def simplify(clause):
-    if isinstance(clause, ast.And):
-        if len(clause.nodes) == 1: result = clause.nodes[0]
-        else: return clause
-    elif isinstance(clause, ast.Or):
-        if len(clause.nodes) == 1: result = ast.Not(clause.nodes[0])
-        else: return clause
-    else: return clause
-    if getattr(result, 'endpos', 0) < clause.endpos: result.endpos = clause.endpos
+    if isinstance(clause, ast.BoolOp) and isinstance(clause.op, ast.And):
+        if len(clause.values) == 1:
+            result = clause.values[0]
+        else:
+            return clause
+    elif isinstance(clause, ast.BoolOp) and isinstance(clause.op, ast.Or):
+        if len(clause.values) == 1:
+            result = ast.UnaryOp(op=ast.Not(), operand=clause.values[0])
+        else:
+            return clause
+    else:
+        return clause
+    if getattr(result, 'endpos', 0) < clause.endpos:
+        result.endpos = clause.endpos
     return result
 
-class InvalidQuery(Exception): pass
 
-def binop(node_type, args_holder=tuple):
+class InvalidQuery(Exception):
+    pass
+
+
+def binop(node_type):
     def method(decompiler):
         oper2 = decompiler.stack.pop()
         oper1 = decompiler.stack.pop()
-        return node_type(args_holder((oper1, oper2)))
+        return ast.BinOp(left=oper1, op=node_type(), right=oper2)
     return method
 
-if not PY2: ord = lambda x: x
+
+operator_mapping = {
+    '==': ast.Eq,
+    '!=': ast.NotEq,
+    '<': ast.Lt,
+    '<=': ast.LtE,
+    '>': ast.Gt,
+    '>=': ast.GtE,
+    'is': ast.Is,
+    'is not': ast.IsNot,
+    'in': ast.In,
+    'not in': ast.NotIn
+}
+
+
+def clean_assign(node):
+    if isinstance(node, ast.Assign):
+        return node.targets
+    return node
+
 
 class Decompiler(object):
     def __init__(decompiler, code, start=0, end=None):
@@ -81,7 +112,8 @@ class Decompiler(object):
         decompiler.decompile()
         decompiler.ast = decompiler.stack.pop()
         decompiler.external_names = decompiler.names - decompiler.assnames
-        assert not decompiler.stack, decompiler.stack
+        if decompiler.stack:
+            throw(DecompileError, 'Compiled code should represent a single expression')
     def get_instructions(decompiler):
         PY36 = sys.version_info >= (3, 6)
         before_yield = True
@@ -91,26 +123,26 @@ class Decompiler(object):
         decompiler.abs_jump_to_top = decompiler.for_iter_pos = -1
         while decompiler.pos < decompiler.end:
             i = decompiler.pos
-            op = ord(code.co_code[i])
+            op = code.co_code[i]
             if PY36:
                 extended_arg = 0
-                oparg = ord(code.co_code[i+1])
+                oparg = code.co_code[i+1]
                 while op == EXTENDED_ARG:
                     extended_arg = (extended_arg | oparg) << 8
                     i += 2
-                    op = ord(code.co_code[i])
-                    oparg = ord(code.co_code[i+1])
+                    op = code.co_code[i]
+                    oparg = code.co_code[i+1]
                 oparg = None if op < HAVE_ARGUMENT else oparg | extended_arg
                 i += 2
             else:
                 i += 1
                 if op >= HAVE_ARGUMENT:
-                    oparg = ord(co_code[i]) + ord(co_code[i + 1]) * 256
+                    oparg = co_code[i] + co_code[i + 1] * 256
                     i += 2
                     if op == EXTENDED_ARG:
-                        op = ord(code.co_code[i])
+                        op = code.co_code[i]
                         i += 1
-                        oparg = ord(co_code[i]) + ord(co_code[i + 1]) * 256 + oparg * 65536
+                        oparg = co_code[i] + co_code[i + 1] * 256 + oparg * 65536
                         i += 2
             if op >= HAVE_ARGUMENT:
                 if op in hasconst: arg = [code.co_consts[oparg]]
@@ -130,6 +162,8 @@ class Decompiler(object):
             if before_yield:
                 if 'JUMP' in opname:
                     endpos = arg[0]
+                    if PY310:
+                        endpos *= 2
                     if endpos < decompiler.pos:
                         decompiler.conditions_end = i
                     decompiler.jump_map[endpos].append(decompiler.pos)
@@ -176,76 +210,85 @@ class Decompiler(object):
                 decompiler.stack.append(x)
 
     def pop_items(decompiler, size):
-        if not size: return ()
+        if not size: return []
         result = decompiler.stack[-size:]
         decompiler.stack[-size:] = []
         return result
     def store(decompiler, node):
         stack = decompiler.stack
-        if not stack: stack.append(node); return
+        if not stack:
+            stack.append(node); return
         top = stack[-1]
-        if isinstance(top, (ast.AssTuple, ast.AssList)) and len(top.nodes) < top.count:
-            top.nodes.append(node)
-            if len(top.nodes) == top.count: decompiler.store(stack.pop())
-        elif isinstance(top, ast.GenExprFor):
-            assert top.assign is None
-            top.assign = node
-        else: stack.append(node)
+        if isinstance(top, ast.Assign):
+            target = top.targets
+            if isinstance(target, (ast.Tuple, ast.List)) and len(target.elts) < top.count:
+                target.elts.append(clean_assign(node))
+                if len(target.elts) == top.count:
+                    decompiler.store(stack.pop())
+            else:
+                stack.append(node)
+        elif isinstance(top, ast.comprehension):
+            assert top.target is None
+            if isinstance(node, ast.Assign):
+                node = node.targets
+            top.target = node
+        else:
+            stack.append(node)
 
-    BINARY_POWER        = binop(ast.Power)
-    BINARY_MULTIPLY     = binop(ast.Mul)
+    BINARY_POWER        = binop(ast.Pow)
+    BINARY_MULTIPLY     = binop(ast.Mult)
     BINARY_DIVIDE       = binop(ast.Div)
     BINARY_FLOOR_DIVIDE = binop(ast.FloorDiv)
     BINARY_ADD          = binop(ast.Add)
     BINARY_SUBTRACT     = binop(ast.Sub)
-    BINARY_LSHIFT       = binop(ast.LeftShift)
-    BINARY_RSHIFT       = binop(ast.RightShift)
-    BINARY_AND          = binop(ast.Bitand, list)
-    BINARY_XOR          = binop(ast.Bitxor, list)
-    BINARY_OR           = binop(ast.Bitor, list)
+    BINARY_LSHIFT       = binop(ast.LShift)
+    BINARY_RSHIFT       = binop(ast.RShift)
+    BINARY_AND          = binop(ast.BitAnd)
+    BINARY_XOR          = binop(ast.BitXor)
+    BINARY_OR           = binop(ast.BitOr)
     BINARY_TRUE_DIVIDE  = BINARY_DIVIDE
     BINARY_MODULO       = binop(ast.Mod)
 
     def BINARY_SUBSCR(decompiler):
-        oper2 = decompiler.stack.pop()
-        oper1 = decompiler.stack.pop()
-        if isinstance(oper2, ast.Sliceobj) and len(oper2.nodes) == 2:
-            a, b = oper2.nodes
-            a = None if isinstance(a, ast.Const) and a.value == None else a
-            b = None if isinstance(b, ast.Const) and b.value == None else b
-            return ast.Slice(oper1, 'OP_APPLY', a, b)
-        elif isinstance(oper2, ast.Tuple):
-            return ast.Subscript(oper1, 'OP_APPLY', list(oper2.nodes))
-        else:
-            return ast.Subscript(oper1, 'OP_APPLY', [ oper2 ])
+        node2 = decompiler.stack.pop()
+        node1 = decompiler.stack.pop()
+        if isinstance(node2, ast.Slice):  # and len(node2.nodes) == 2:
+            if isinstance(node2.lower, ast.Constant) and node2.lower.value is None:
+                node2.lower = None
+            if isinstance(node2.upper, ast.Constant) and node2.upper.value is None:
+                node2.upper = None
+        return ast.Subscript(value=node1, slice=node2, ctx=ast.Load())
 
     def BUILD_CONST_KEY_MAP(decompiler, length):
         keys = decompiler.stack.pop()
-        assert isinstance(keys, ast.Const)
-        keys = [ ast.Const(key) for key in keys.value ]
+        assert isinstance(keys, ast.Constant)
+        keys = [ ast.Constant(key) for key in keys.value ]
         values = decompiler.pop_items(length)
-        pairs = list(izip(keys, values))
-        return ast.Dict(pairs)
+        return ast.Dict(keys=keys, values=values)
 
     def BUILD_LIST(decompiler, size):
-        return ast.List(decompiler.pop_items(size))
+        return ast.List(decompiler.pop_items(size), ast.Load())
 
     def BUILD_MAP(decompiler, length):
         if sys.version_info < (3, 5):
             return ast.Dict(())
         data = decompiler.pop_items(2 * length)  # [key1, value1, key2, value2, ...]
-        it = iter(data)
-        pairs = list(izip(it, it))  # [(key1, value1), (key2, value2), ...]
-        return ast.Dict(tuple(pairs))
+        keys, values = [], []
+        for i in range(0, len(data), 2):
+            keys.append(data[i])
+            values.append(data[i+1])
+        return ast.Dict(keys=keys, values=values)
 
     def BUILD_SET(decompiler, size):
         return ast.Set(decompiler.pop_items(size))
 
     def BUILD_SLICE(decompiler, size):
-        return ast.Sliceobj(decompiler.pop_items(size))
+        items = decompiler.pop_items(size)
+        items += [None] * (3 - len(items))
+        return ast.Slice(*items, ctx=ast.Load())
 
     def BUILD_TUPLE(decompiler, size):
-        return ast.Tuple(decompiler.pop_items(size))
+        return ast.Tuple(decompiler.pop_items(size), ast.Load())
 
     def BUILD_STRING(decompiler, count):
         values = list(reversed([decompiler.stack.pop() for _ in range(count)]))
@@ -254,26 +297,33 @@ class Decompiler(object):
     def CALL_FUNCTION(decompiler, argc, star=None, star2=None):
         pop = decompiler.stack.pop
         kwarg, posarg = divmod(argc, 256)
-        args = []
-        for i in xrange(kwarg):
+        keywords = []
+        for i in range(kwarg):
             arg = pop()
             key = pop().value
-            args.append(ast.Keyword(key, arg))
-        for i in xrange(posarg): args.append(pop())
+            keywords.append(ast.keyword(key, arg))
+        keywords.reverse()
+        args = []
+        for i in range(posarg):
+            args.append(pop())
         args.reverse()
-        return decompiler._call_function(args, star, star2)
+        if star:
+            args.append(ast.Starred(value=star))
+        if star2:
+            keywords.append(ast.keyword(value=star2))
+        return decompiler._call_function(args, keywords)
 
-    def _call_function(decompiler, args, star=None, star2=None):
+    def _call_function(decompiler, args, keywords=None):
         tos = decompiler.stack.pop()
-        if isinstance(tos, ast.GenExpr):
-            assert len(args) == 1 and star is None and star2 is None
+        if isinstance(tos, ast.GeneratorExp):
+            assert len(args) == 1 and not keywords
             genexpr = tos
-            qual = genexpr.code.quals[0]
+            qual = genexpr.generators[0]
             assert isinstance(qual.iter, ast.Name)
-            assert qual.iter.name in ('.0', '[outmost-iterable]')
+            assert qual.iter.id == '.0'
             qual.iter = args[0]
             return genexpr
-        else: return ast.CallFunc(tos, args, star, star2)
+        return ast.Call(tos, args, keywords)
 
     def CALL_FUNCTION_VAR(decompiler, argc):
         return decompiler.CALL_FUNCTION(argc, decompiler.stack.pop())
@@ -282,14 +332,13 @@ class Decompiler(object):
         if sys.version_info < (3, 6):
             return decompiler.CALL_FUNCTION(argc, star2=decompiler.stack.pop())
         keys = decompiler.stack.pop()
-        assert isinstance(keys, ast.Const)
+        assert isinstance(keys, ast.Constant)
         keys = keys.value
         values = decompiler.pop_items(argc)
         assert len(keys) <= len(values)
         args = values[:-len(keys)]
-        for key, value in izip(keys, values[-len(keys):]):
-            args.append(ast.Keyword(key, value))
-        return decompiler._call_function(args)
+        keywords = [ast.keyword(k, v) for k, v in zip(keys, values[-len(keys):])]
+        return decompiler._call_function(args, keywords)
 
     def CALL_FUNCTION_VAR_KW(decompiler, argc):
         star2 = decompiler.stack.pop()
@@ -299,33 +348,38 @@ class Decompiler(object):
     def CALL_FUNCTION_EX(decompiler, argc):
         star2 = None
         if argc:
-            if argc != 1: throw(DecompileError)
+            if argc != 1:
+                throw(DecompileError)
             star2 = decompiler.stack.pop()
         star = decompiler.stack.pop()
-        return decompiler._call_function([], star, star2)
+        args = [ast.Starred(value=star)] if star else None
+        keywords = [ast.keyword(value=star2)] if star2 else None
+        return decompiler._call_function(args, keywords)
 
     def CALL_METHOD(decompiler, argc):
         pop = decompiler.stack.pop
         args = []
+        keywords = []
         if argc >= 256:
             kwargc = argc // 256
             argc = argc % 256
             for i in range(kwargc):
                 v = pop()
                 k = pop()
-                assert isinstance(k, ast.Const)
-                k = k.value # ast.Name(k.value)
-                args.append(ast.Keyword(k, v))
+                assert isinstance(k, ast.Constant)
+                k = k.value  # ast.Name(k.value)
+                keywords.append(ast.keyword(k, v))
         for i in range(argc):
             args.append(pop())
         args.reverse()
         method = pop()
-        return ast.CallFunc(method, args)
+        return ast.Call(method, args, keywords)
 
     def COMPARE_OP(decompiler, op):
         oper2 = decompiler.stack.pop()
         oper1 = decompiler.stack.pop()
-        return ast.Compare(oper1, [(op, oper2)])
+        op = operator_mapping[op]()
+        return ast.Compare(oper1, [op], [oper2])
 
     def CONTAINS_OP(decompiler, invert):
         return decompiler.COMPARE_OP('not in' if invert else 'in')
@@ -334,19 +388,30 @@ class Decompiler(object):
         return decompiler.stack[-1]
 
     def FOR_ITER(decompiler, endpos):
-        assign = None
+        target = None
         iter = decompiler.stack.pop()
         ifs = []
-        return ast.GenExprFor(assign, iter, ifs)
+        return ast.comprehension(target, iter, ifs, 0)
 
     def FORMAT_VALUE(decompiler, flags):
         if flags in (0, 1, 2, 3):
             value = decompiler.stack.pop()
-            return ast.Str(value, flags)
+            if flags == 0:
+                return value
+            elif flags == 1:
+                conversion = ord('s')  # str conversion
+            elif flags == 2:
+                conversion = ord('r')  # repr conversion
+            elif flags == 3:
+                conversion = ord('a')  # ascii conversion
+            return ast.FormattedValue(value=value, conversion=conversion)
         elif flags == 4:
             fmt_spec = decompiler.stack.pop()
             value = decompiler.stack.pop()
-            return ast.FormattedValue(value, fmt_spec)
+            return ast.FormattedValue(value=value, format_spec=fmt_spec)
+
+    def GEN_START(decompiler, kind):
+        assert kind == 0  # only support sync
 
     def GET_ITER(decompiler):
         pass
@@ -375,57 +440,72 @@ class Decompiler(object):
             decompiler.process_target(i)
         expr = decompiler.stack.pop()
         clausetype = ast.Or if if_true else ast.And
-        clause = clausetype([expr])
+        clause = ast.BoolOp(op=clausetype(), values=[expr])
         clause.endpos = endpos
         decompiler.targets.setdefault(endpos, clause)
         return clause
 
     def conditional_jump_new(decompiler, endpos, if_true):
+        if PY310:
+            endpos *= 2
         expr = decompiler.stack.pop()
         if decompiler.pos >= decompiler.conditions_end:
             clausetype = ast.Or if if_true else ast.And
         elif decompiler.pos in decompiler.or_jumps:
             clausetype = ast.Or
             if not if_true:
-                expr = ast.Not(expr)
+                expr = ast.UnaryOp(op=ast.Not(), operand=expr)
         else:
             clausetype = ast.And
             if if_true:
-                expr = ast.Not(expr)
+                expr = ast.UnaryOp(op=ast.Not(), operand=expr)
         decompiler.stack.append(expr)
 
         if decompiler.next_pos in decompiler.targets:
             decompiler.process_target(decompiler.next_pos)
 
         expr = decompiler.stack.pop()
-        clause = clausetype([ expr ])
+        clause = ast.BoolOp(op=clausetype(), values=[expr])
         clause.endpos = endpos
         decompiler.targets.setdefault(endpos, clause)
         return clause
 
     def process_target(decompiler, pos, partial=False):
-        if pos is None: limit = None
-        elif partial: limit = decompiler.targets.get(pos, None)
-        else: limit = decompiler.targets.pop(pos, None)
+        if pos is None:
+            limit = None
+        elif partial:
+            limit = decompiler.targets.get(pos, None)
+        else:
+            limit = decompiler.targets.pop(pos, None)
         top = decompiler.stack.pop()
         while True:
             top = simplify(top)
-            if top is limit: break
-            if isinstance(top, ast.GenExprFor): break
-            if not decompiler.stack: break
+            if top is limit:
+                break
+            if isinstance(top, ast.comprehension):
+                break
+            if not decompiler.stack:
+                break
             top2 = decompiler.stack[-1]
-            if isinstance(top2, ast.GenExprFor): break
-            if partial and hasattr(top2, 'endpos') and top2.endpos == pos: break
+            if isinstance(top2, ast.comprehension):
+                break
+            if partial and hasattr(top2, 'endpos') and top2.endpos == pos:
+                break
 
-            if isinstance(top2, (ast.And, ast.Or)):
-                if top2.__class__ == top.__class__: top2.nodes.extend(top.nodes)
-                else: top2.nodes.append(top)
+            if isinstance(top2, ast.BoolOp):
+                if isinstance(top, ast.BoolOp) and type(top2.op) is type(top.op):
+                    top2.values.extend(top.values)
+                else:
+                    top2.values.append(top)
             elif isinstance(top2, ast.IfExp):  # Python 2.5
                 top2.else_ = top
                 if hasattr(top, 'endpos'):
                     top2.endpos = top.endpos
-                    if decompiler.targets.get(top.endpos) is top: decompiler.targets[top.endpos] = top2
-            else: throw(DecompileError('Expression is too complex to decompile, try to pass query as string, e.g. select("x for x in Something")'))
+                    if decompiler.targets.get(top.endpos) is top:
+                        decompiler.targets[top.endpos] = top2
+            else:
+                throw(DecompileError('Expression is too complex to decompile, try to pass query as string, '
+                                     'e.g. select("x for x in Something")'))
             top2.endpos = max(top2.endpos, getattr(top, 'endpos', 0))
             top = decompiler.stack.pop()
         decompiler.stack.append(top)
@@ -436,10 +516,11 @@ class Decompiler(object):
         then = decompiler.stack.pop()
         decompiler.process_target(i, False)
         test = decompiler.stack.pop()
-        if_exp = ast.IfExp(simplify(test), simplify(then), None)
+        if_exp = ast.IfExp(test=simplify(test), body=simplify(then), orelse=None)
         if_exp.endpos = endpos
         decompiler.targets.setdefault(endpos, if_exp)
-        if decompiler.targets.get(endpos) is then: decompiler.targets[endpos] = if_exp
+        if decompiler.targets.get(endpos) is then:
+            decompiler.targets[endpos] = if_exp
         return if_exp
 
     def IS_OP(decompiler, invert):
@@ -453,38 +534,38 @@ class Decompiler(object):
         if offset != 1:
             raise NotImplementedError(offset)
         items = decompiler.stack.pop()
-        if not isinstance(items, ast.Const):
+        if not isinstance(items, ast.Constant):
             raise NotImplementedError(type(items))
         if not isinstance(items.value, tuple):
             raise NotImplementedError(type(items.value))
         lst = decompiler.stack.pop()
         if not isinstance(lst, ast.List):
             raise NotImplementedError(type(lst))
-        values = tuple(ast.Const(v) for v in items.value)
-        lst.nodes = lst.nodes + values
+        values = [ast.Constant(v) for v in items.value]
+        lst.elts.extend(values)
         return lst
 
     def LOAD_ATTR(decompiler, attr_name):
-        return ast.Getattr(decompiler.stack.pop(), attr_name)
+        return ast.Attribute(decompiler.stack.pop(), attr_name, ast.Load())
 
     def LOAD_CLOSURE(decompiler, freevar):
         decompiler.names.add(freevar)
-        return ast.Name(freevar)
+        return ast.Name(freevar, ast.Load())
 
     def LOAD_CONST(decompiler, const_value):
-        return ast.Const(const_value)
+        return ast.Constant(const_value)
 
     def LOAD_DEREF(decompiler, freevar):
         decompiler.names.add(freevar)
-        return ast.Name(freevar)
+        return ast.Name(freevar, ast.Load())
 
     def LOAD_FAST(decompiler, varname):
         decompiler.names.add(varname)
-        return ast.Name(varname)
+        return ast.Name(varname, ast.Load())
 
     def LOAD_GLOBAL(decompiler, varname):
         decompiler.names.add(varname)
-        return ast.Name(varname)
+        return ast.Name(varname, ast.Load())
 
     def LOAD_METHOD(decompiler, methname):
         return decompiler.LOAD_ATTR(methname)
@@ -493,16 +574,14 @@ class Decompiler(object):
 
     def LOAD_NAME(decompiler, varname):
         decompiler.names.add(varname)
-        return ast.Name(varname)
+        return ast.Name(varname, ast.Load())
 
     def MAKE_CLOSURE(decompiler, argc):
-        if PY2: decompiler.stack[-2:-1] = [] # ignore freevars
-        else: decompiler.stack[-3:-2] = [] # ignore freevars
+        decompiler.stack[-3:-2] = []  # ignore freevars
         return decompiler.MAKE_FUNCTION(argc)
 
     def MAKE_FUNCTION(decompiler, argc):
         defaults = []
-        flags = 0
         if sys.version_info >= (3, 6):
             qualname = decompiler.stack.pop()
             tos = decompiler.stack.pop()
@@ -516,8 +595,7 @@ class Decompiler(object):
                 defaults = decompiler.stack.pop()
                 throw(DecompileError)
         else:
-            if not PY2:
-                qualname = decompiler.stack.pop()
+            qualname = decompiler.stack.pop()
             tos = decompiler.stack.pop()
             if argc:
                 defaults = [ decompiler.stack.pop() for i in range(argc) ]
@@ -530,11 +608,17 @@ class Decompiler(object):
         argnames, varargs, keywords = inspect.getargs(codeobject)
         if varargs:
             argnames.append(varargs)
-            flags |= inspect.CO_VARARGS
         if keywords:
             argnames.append(keywords)
-            flags |= inspect.CO_VARKEYWORDS
-        return ast.Lambda(argnames, defaults, flags, func_decompiler.ast)
+        args = ast.arguments(
+            posonlyargs=[],
+            args=[ast.arg(arg=v) for v in argnames],
+            kwonlyargs=[],
+            kw_defaults=[],
+            defaults=defaults,
+            kwarg=None
+        )
+        return ast.Lambda(args, func_decompiler.ast)
 
     POP_JUMP_IF_FALSE = JUMP_IF_FALSE
     POP_JUMP_IF_TRUE = JUMP_IF_TRUE
@@ -543,7 +627,8 @@ class Decompiler(object):
         pass
 
     def RETURN_VALUE(decompiler):
-        if decompiler.next_pos != decompiler.end: throw(DecompileError)
+        if decompiler.next_pos != decompiler.end:
+            throw(DecompileError)
         expr = decompiler.stack.pop()
         return simplify(expr)
 
@@ -564,90 +649,73 @@ class Decompiler(object):
     def SETUP_LOOP(decompiler, endpos):
         pass
 
-    def SLICE_0(decompiler):
-        return ast.Slice(decompiler.stack.pop(), 'OP_APPLY', None, None)
-
-    def SLICE_1(decompiler):
-        tos = decompiler.stack.pop()
-        tos1 = decompiler.stack.pop()
-        return ast.Slice(tos1, 'OP_APPLY', tos, None)
-
-    def SLICE_2(decompiler):
-        tos = decompiler.stack.pop()
-        tos1 = decompiler.stack.pop()
-        return ast.Slice(tos1, 'OP_APPLY', None, tos)
-
-    def SLICE_3(decompiler):
-        tos = decompiler.stack.pop()
-        tos1 = decompiler.stack.pop()
-        tos2 = decompiler.stack.pop()
-        return ast.Slice(tos2, 'OP_APPLY', tos1, tos)
-
     def STORE_ATTR(decompiler, attrname):
-        decompiler.store(ast.AssAttr(decompiler.stack.pop(), attrname, 'OP_ASSIGN'))
+        decompiler.store(ast.Attribute(decompiler.stack.pop(), attrname, ast.Store()))
 
     def STORE_DEREF(decompiler, freevar):
         decompiler.assnames.add(freevar)
-        decompiler.store(ast.AssName(freevar, 'OP_ASSIGN'))
+        decompiler.store(ast.Name(freevar, ast.Store()))
 
     def STORE_FAST(decompiler, varname):
         if varname.startswith('_['):
             throw(InvalidQuery('Use generator expression (... for ... in ...) '
                                'instead of list comprehension [... for ... in ...] inside query'))
         decompiler.assnames.add(varname)
-        decompiler.store(ast.AssName(varname, 'OP_ASSIGN'))
+        decompiler.store(ast.Name(varname, ast.Store()))
 
     def STORE_MAP(decompiler):
         tos = decompiler.stack.pop()
         tos1 = decompiler.stack.pop()
         tos2 = decompiler.stack[-1]
-        if not isinstance(tos2, ast.Dict): assert False  # pragma: no cover
-        if tos2.items == (): tos2.items = []
+        if not isinstance(tos2, ast.Dict):
+            assert False  # pragma: no cover
+        if tos2.items == ():
+            tos2.items = []
         tos2.items.append((tos, tos1))
 
     def STORE_SUBSCR(decompiler):
         tos = decompiler.stack.pop()
         tos1 = decompiler.stack.pop()
         tos2 = decompiler.stack.pop()
-        if not isinstance(tos1, ast.Dict): assert False  # pragma: no cover
-        if tos1.items == (): tos1.items = []
+        if not isinstance(tos1, ast.Dict):
+            assert False  # pragma: no cover
+        if tos1.items == ():
+            tos1.items = []
         tos1.items.append((tos, tos2))
 
     def UNARY_POSITIVE(decompiler):
-        return ast.UnaryAdd(decompiler.stack.pop())
+        return ast.UnaryOp(op=ast.UAdd(), operand=decompiler.stack.pop())
 
     def UNARY_NEGATIVE(decompiler):
-        return ast.UnarySub(decompiler.stack.pop())
+        return ast.UnaryOp(op=ast.USub(), operand=decompiler.stack.pop())
 
     def UNARY_NOT(decompiler):
-        return ast.Not(decompiler.stack.pop())
-
-    def UNARY_CONVERT(decompiler):
-        return ast.Backquote(decompiler.stack.pop())
+        return ast.UnaryOp(op=ast.Not(), operand=decompiler.stack.pop())
 
     def UNARY_INVERT(decompiler):
         return ast.Invert(decompiler.stack.pop())
 
     def UNPACK_SEQUENCE(decompiler, count):
-        ass_tuple = ast.AssTuple([])
+        ass_tuple = ast.Assign(targets=ast.Tuple([], ast.Store()))
         ass_tuple.count = count
         return ass_tuple
 
     def YIELD_VALUE(decompiler):
         expr = decompiler.stack.pop()
-        fors = []
+        generators = []
         while decompiler.stack:
             decompiler.process_target(None)
             top = decompiler.stack.pop()
-            if not isinstance(top, (ast.GenExprFor)):
-                cond = ast.GenExprIf(top)
+            if not isinstance(top, ast.comprehension):
+                cond = top
                 top = decompiler.stack.pop()
-                assert isinstance(top, ast.GenExprFor)
+                assert isinstance(top, ast.comprehension)
                 top.ifs.append(cond)
-                fors.append(top)
-            else: fors.append(top)
-        fors.reverse()
-        return ast.GenExpr(ast.GenExprInner(simplify(expr), fors))
+                generators.append(top)
+            else:
+                generators.append(top)
+        generators.reverse()
+        return ast.GeneratorExp(simplify(expr), generators)
 
 test_lines = """
     (a and b if c and d else e and f for i in T if (A and B if C and D else E and F))
@@ -736,25 +804,30 @@ def test():
     if sys.version[:3] > '2.4': outmost_iterable_name = '.0'
     else: outmost_iterable_name = '[outmost-iterable]'
     import dis
-    for line in test_lines.split('\n'):
-        if not line or line.isspace(): continue
+    for i, line in enumerate(test_lines.split('\n')):
+        if not line or line.isspace():
+            continue
         line = line.strip()
-        if line.startswith('#'): continue
+        if line.startswith('#'):
+            continue
         code = compile(line, '<?>', 'eval').co_consts[0]
-        ast1 = parse(line).node.nodes[0].expr
-        ast1.code.quals[0].iter.name = outmost_iterable_name
-        try: ast2 = Decompiler(code).ast
+        ast1 = ast.parse(line).body[0]
+        ast1.value.generators[0].iter.id = outmost_iterable_name
+        ast1 = ast.dump(ast1, indent=2)
+        try:
+            ast2 = ast.Expr(Decompiler(code).ast)
+            ast2 = ast.dump(ast2, indent=2)
         except Exception as e:
             print()
-            print(line)
+            print(i, line)
             print()
             print(ast1)
             print()
             dis.dis(code)
             raise
-        if str(ast1) != str(ast2):
+        if ast1 != ast2:
             print()
-            print(line)
+            print(i, line)
             print()
             print(ast1)
             print()
@@ -762,7 +835,8 @@ def test():
             print()
             dis.dis(code)
             break
-        else: print('OK: %s' % line)
+        else: print('%d OK: %s' % (i, line))
     else: print('Done!')
 
-if __name__ == '__main__': test()
+if __name__ == '__main__':
+    test()
